@@ -1,32 +1,38 @@
 package Zonemaster::Engine;
 
-use version; our $VERSION = version->declare("v2.0.7");
+use version; our $VERSION = version->declare("v3.1.2");
 
 use 5.014002;
-use Moose;
 
+BEGIN {
+    # Locale::TextDomain (<= 1.20) doesn't know about File::ShareDir so give a helping hand.
+    # This is a hugely simplified version of the reference implementation located here:
+    # https://metacpan.org/source/GUIDO/libintl-perl-1.21/lib/Locale/TextDomain.pm
+    require File::ShareDir;
+    require Locale::TextDomain;
+    my $share = File::ShareDir::dist_dir( 'Zonemaster-Engine' );
+    Locale::TextDomain->import( 'Zonemaster-Engine', "$share/locale" );
+}
+
+use Moose;
+use Carp;
 use Zonemaster::Engine::Nameserver;
 use Zonemaster::Engine::Logger;
-use Zonemaster::Engine::Config;
+use Zonemaster::Engine::Profile;
 use Zonemaster::Engine::Zone;
 use Zonemaster::Engine::Test;
 use Zonemaster::Engine::Recursor;
 use Zonemaster::Engine::ASNLookup;
 
 our $logger;
-our $config;
 our $recursor = Zonemaster::Engine::Recursor->new;
 
 sub logger {
     return $logger //= Zonemaster::Engine::Logger->new;
 }
 
-sub config {
-    if ( not defined $config ) {
-        $config = Zonemaster::Engine::Config->new;
-    }
-
-    return $config;
+sub profile {
+    return Zonemaster::Engine::Profile->effective;
 }
 
 sub ns {
@@ -99,9 +105,17 @@ sub recurse {
 
 sub add_fake_delegation {
     my ( $class, $domain, $href ) = @_;
-    my $incomplete_delegation;
+
+    # Validate arguments
+    $domain =~ /[^.]$|^\.$/
+      or croak 'Argument $domain must omit the trailing dot, or it must be a single dot';
+    foreach my $name ( keys %{$href} ) {
+        $name =~ /[^.]$|^\.$/
+          or croak 'Each key of argument $href must omit the trailing dot, or it must be a single dot';
+    }
 
     # Check fake delegation
+    my $incomplete_delegation;
     foreach my $name ( keys %{$href} ) {
         if ( not defined $href->{$name} or not scalar @{ $href->{$name} } ) {
             if ( Zonemaster::Engine::Zone->new( { name => $domain } )->is_in_zone( $name ) ) {
@@ -122,7 +136,7 @@ sub add_fake_delegation {
                     );
                     push @{ $href->{$name} }, ();
                     $incomplete_delegation = 1;
-		}
+                }
             }
         }
     }
@@ -233,9 +247,9 @@ Returns a L<Zonemaster::Engine::Zone> object for the given name.
 
 Returns a L<Zonemaster::Engine::Nameserver> object for the given name and address.
 
-=item config()
+=item profile()
 
-Returns the global L<Zonemaster::Engine::Config> object.
+Returns the effective profile (L<Zonemaster::Engine::Profile> object).
 
 =item logger()
 
