@@ -1,6 +1,6 @@
 package Zonemaster::Engine::Test::Consistency;
 
-use version; our $VERSION = version->declare("v1.1.2");
+use version; our $VERSION = version->declare("v1.1.12");
 
 use strict;
 use warnings;
@@ -8,11 +8,13 @@ use warnings;
 use 5.014002;
 
 use Zonemaster::Engine;
-use Zonemaster::Engine::Util;
-use Zonemaster::Engine::Test::Address;
-use Zonemaster::Engine::Constants qw[:ip :soa];
 
 use List::MoreUtils qw[uniq];
+use Locale::TextDomain qw[Zonemaster-Engine];
+use Readonly;
+use Zonemaster::Engine::Constants qw[:ip :soa];
+use Zonemaster::Engine::Test::Address;
+use Zonemaster::Engine::Util;
 
 ###
 ### Entry points
@@ -22,21 +24,24 @@ sub all {
     my ( $class, $zone ) = @_;
     my @results;
 
-    if ( Zonemaster::Engine->config->should_run( 'consistency01' ) ) {
+    if ( Zonemaster::Engine::Util::should_run_test( q{consistency01} ) ) {
         push @results, $class->consistency01( $zone );
     }
-    if ( Zonemaster::Engine->config->should_run( 'consistency02' ) ) {
+    if ( Zonemaster::Engine::Util::should_run_test( q{consistency02} ) ) {
         push @results, $class->consistency02( $zone );
     }
-    if ( Zonemaster::Engine->config->should_run( 'consistency03' ) ) {
+    if ( Zonemaster::Engine::Util::should_run_test( q{consistency03} ) ) {
         push @results, $class->consistency03( $zone );
     }
-    if ( Zonemaster::Engine->config->should_run( 'consistency04' ) ) {
+    if ( Zonemaster::Engine::Util::should_run_test( q{consistency04} ) ) {
         push @results, $class->consistency04( $zone );
     }
-    if ( Zonemaster::Engine->config->should_run( 'consistency05' ) ) {
+    if ( Zonemaster::Engine::Util::should_run_test( q{consistency05} ) ) {
         push @results, $class->consistency05( $zone );
     }
+    if ( Zonemaster::Engine::Util::should_run_test( q{consistency06} ) ) {
+        push @results, $class->consistency06( $zone );
+    }
 
     return @results;
 }
@@ -96,44 +101,152 @@ sub metadata {
         ],
         consistency05 => [
             qw(
-              EXTRA_ADDRESS_PARENT
-              EXTRA_ADDRESS_CHILD
-              TOTAL_ADDRESS_MISMATCH
               ADDRESSES_MATCH
+              CHILD_NS_FAILED
+              CHILD_ZONE_LAME
+              EXTRA_ADDRESS_CHILD
+              IN_BAILIWICK_ADDR_MISMATCH
+              NO_RESPONSE
+              OUT_OF_BAILIWICK_ADDR_MISMATCH
+              )
+        ],
+        consistency06 => [
+            qw(
+              NO_RESPONSE
+              NO_RESPONSE_SOA_QUERY
+              ONE_SOA_MNAME
+              MULTIPLE_SOA_MNAMES
               )
         ],
     };
 } ## end sub metadata
 
-sub translation {
-    return {
-        'SOA_TIME_PARAMETER_SET' =>
-'Saw SOA time parameter set (REFRESH={refresh},RETRY={retry},EXPIRE={expire},MINIMUM={minimum}) on following nameserver set : {servers}.',
-        'ONE_SOA_RNAME'                   => 'A single SOA rname value was seen ({rname})',
-        'MULTIPLE_SOA_SERIALS'            => 'Saw {count} SOA serial numbers.',
-        'SOA_SERIAL'                      => 'Saw SOA serial number {serial} on following nameserver set : {servers}.',
-        'SOA_RNAME'                       => 'Saw SOA rname {rname} on following nameserver set : {servers}.',
-        'MULTIPLE_SOA_RNAMES'             => 'Saw {count} SOA rname.',
-        'ONE_SOA_SERIAL'                  => 'A single SOA serial number was seen ({serial}).',
-        'MULTIPLE_SOA_TIME_PARAMETER_SET' => 'Saw {count} SOA time parameter set.',
-        'NO_RESPONSE'                     => 'Nameserver {ns}/{address} did not respond.',
-        'ONE_SOA_TIME_PARAMETER_SET' =>
-'A single SOA time parameter set was seen (REFRESH={refresh},RETRY={retry},EXPIRE={expire},MINIMUM={minimum}).',
-        'NO_RESPONSE_SOA_QUERY' => 'No response from nameserver {ns}/{address} on SOA queries.',
-        'SOA_SERIAL_VARIATION' =>
-'Difference between the smaller serial ({serial_min}) and the bigger one ({serial_max}) is greater than the maximum allowed ({max_variation}).',
-        'NO_RESPONSE_NS_QUERY' => 'No response from nameserver {ns}/{address} on NS queries.',
-        'ONE_NS_SET'           => 'A unique NS set was seen ({nsset}).',
-        'MULTIPLE_NS_SET'      => 'Saw {count} NS set.',
-        'NS_SET'               => 'Saw NS set ({nsset}) on following nameserver set : {servers}.',
-        'IPV4_DISABLED'        => 'IPv4 is disabled, not sending "{rrtype}" query to {ns}/{address}.',
-        'IPV6_DISABLED'        => 'IPv6 is disabled, not sending "{rrtype}" query to {ns}/{address}.',
-        'EXTRA_ADDRESS_PARENT' => 'Parent has extra nameserver IP address(es) not listed at child ({addresses}).',
-        'EXTRA_ADDRESS_CHILD'  => 'Child has extra nameserver IP address(es) not listed at parent ({addresses}).',
-        'TOTAL_ADDRESS_MISMATCH' => 'No common nameserver IP addresses between child ({child}) and parent ({glue}).',
-        'ADDRESSES_MATCH'        => 'Glue records are consistent between glue and authoritative data.',
-    };
-} ## end sub translation
+Readonly my %TAG_DESCRIPTIONS => (
+    ADDRESSES_MATCH => sub {
+        __x    # CONSISTENCY:ADDRESSES_MATCH
+          'Glue records are consistent between glue and authoritative data.', @_;
+    },
+    CHILD_NS_FAILED => sub {
+        __x    # CONSISTENCY:CHILD_NS_FAILED
+          'Unexpected or erroneous reply from {ns}/{address}.', @_;
+    },
+    CHILD_ZONE_LAME => sub {
+        __x    # CONSISTENCY:CHILD_ZONE_LAME
+          'Lame delegation', @_;
+    },
+    EXTRA_ADDRESS_CHILD => sub {
+        __x    # CONSISTENCY:EXTRA_ADDRESS_CHILD
+          'Child has extra nameserver IP address(es) not listed at parent ({addresses}).', @_;
+    },
+    EXTRA_ADDRESS_PARENT => sub {
+        __x    # CONSISTENCY:EXTRA_ADDRESS_PARENT
+          'Parent has extra nameserver IP address(es) not listed at child ({addresses}).', @_;
+    },
+    IN_BAILIWICK_ADDR_MISMATCH => sub {
+        __x    # CONSISTENCY:IN_BAILIWICK_ADDR_MISMATCH
+          'In-bailiwick name server listed at parent has a mismatch between glue data at parent '
+          . '({parent_addresses}) and any equivalent address record in child zone ({zone_addresses})',
+          @_;
+    },
+    IPV4_DISABLED => sub {
+        __x    # CONSISTENCY:IPV4_DISABLED
+          'IPv4 is disabled, not sending "{rrtype}" query to {ns}/{address}.', @_;
+    },
+    IPV6_DISABLED => sub {
+        __x    # CONSISTENCY:IPV6_DISABLED
+          'IPv6 is disabled, not sending "{rrtype}" query to {ns}/{address}.', @_;
+    },
+    MULTIPLE_NS_SET => sub {
+        __x    # CONSISTENCY:MULTIPLE_NS_SET
+          'Saw {count} NS set.', @_;
+    },
+    MULTIPLE_SOA_MNAMES => sub {
+        __x    # CONSISTENCY:MULTIPLE_SOA_MNAMES
+          'Saw {count} SOA mname.', @_;
+    },
+    MULTIPLE_SOA_RNAMES => sub {
+        __x    # CONSISTENCY:MULTIPLE_SOA_RNAMES
+          'Saw {count} SOA rname.', @_;
+    },
+    MULTIPLE_SOA_SERIALS => sub {
+        __x    # CONSISTENCY:MULTIPLE_SOA_SERIALS
+          'Saw {count} SOA serial numbers.', @_;
+    },
+    MULTIPLE_SOA_TIME_PARAMETER_SET => sub {
+        __x    # CONSISTENCY:MULTIPLE_SOA_TIME_PARAMETER_SET
+          'Saw {count} SOA time parameter set.', @_;
+    },
+    NO_RESPONSE => sub {
+        __x    # CONSISTENCY:NO_RESPONSE
+          'Nameserver {ns}/{address} did not respond.', @_;
+    },
+    NO_RESPONSE_NS_QUERY => sub {
+        __x    # CONSISTENCY:NO_RESPONSE_NS_QUERY
+          'No response from nameserver {ns}/{address} on NS queries.', @_;
+    },
+    NO_RESPONSE_SOA_QUERY => sub {
+        __x    # CONSISTENCY:NO_RESPONSE_SOA_QUERY
+          'No response from nameserver {ns}/{address} on SOA queries.', @_;
+    },
+    NS_SET => sub {
+        __x    # CONSISTENCY:NS_SET
+          'Saw NS set ({nsset}) on following nameserver set : {servers}.', @_;
+    },
+    ONE_NS_SET => sub {
+        __x    # CONSISTENCY:ONE_NS_SET
+          'A unique NS set was seen ({nsset}).', @_;
+    },
+    ONE_SOA_MNAME => sub {
+        __x    # CONSISTENCY:ONE_SOA_MNAME
+          'A single SOA mname value was seen ({mname})', @_;
+    },
+    ONE_SOA_RNAME => sub {
+        __x    # CONSISTENCY:ONE_SOA_RNAME
+          'A single SOA rname value was seen ({rname})', @_;
+    },
+    ONE_SOA_SERIAL => sub {
+        __x    # CONSISTENCY:ONE_SOA_SERIAL
+          'A single SOA serial number was seen ({serial}).', @_;
+    },
+    ONE_SOA_TIME_PARAMETER_SET => sub {
+        __x    # CONSISTENCY:ONE_SOA_TIME_PARAMETER_SET
+          'A single SOA time parameter set was seen '
+          . '(REFRESH={refresh},RETRY={retry},EXPIRE={expire},MINIMUM={minimum}).',
+          @_;
+    },
+    OUT_OF_BAILIWICK_ADDR_MISMATCH => sub {
+        __x    # CONSISTENCY:OUT_OF_BAILIWICK_ADDR_MISMATCH
+          'Out-of-bailiwick name server listed at parent with glue record has a mismatch between '
+          . 'the glue at the parent ({parent_addresses}) and any equivalent address record found '
+          . 'in authoritative zone ({zone_addresses})', @_;
+    },
+    SOA_RNAME => sub {
+        __x    # CONSISTENCY:SOA_RNAME
+          'Saw SOA rname {rname} on following nameserver set : {servers}.', @_;
+    },
+    SOA_SERIAL => sub {
+        __x    # CONSISTENCY:SOA_SERIAL
+          'Saw SOA serial number {serial} on following nameserver set : {servers}.', @_;
+    },
+    SOA_SERIAL_VARIATION => sub {
+        __x    # CONSISTENCY:SOA_SERIAL_VARIATION
+          'Difference between the smaller serial ({serial_min}) and the bigger one ({serial_max}) '
+          . 'is greater than the maximum allowed ({max_variation}).', @_;
+    },
+    SOA_TIME_PARAMETER_SET => sub {
+        __x    # CONSISTENCY:SOA_TIME_PARAMETER_SET
+          'Saw SOA time parameter set (REFRESH={refresh},RETRY={retry},EXPIRE={expire},'
+          . 'MINIMUM={minimum}) on following nameserver set : {servers}.', @_;
+    },
+    TOTAL_ADDRESS_MISMATCH => sub {
+        __x    # CONSISTENCY:TOTAL_ADDRESS_MISMATCH
+          'No common nameserver IP addresses between child ({child}) and parent ({glue}).', @_;
+    },
+);
+
+sub tag_descriptions {
+    return \%TAG_DESCRIPTIONS;
+}
 
 sub version {
     return "$Zonemaster::Engine::Test::Consistency::VERSION";
@@ -156,7 +269,7 @@ sub consistency01 {
 
         next if $nsnames_and_ip{ $local_ns->name->string . q{/} . $local_ns->address->short };
 
-        if ( not Zonemaster::Engine->config->ipv6_ok and $local_ns->address->version == $IP_VERSION_6 ) {
+        if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv6}) and $local_ns->address->version == $IP_VERSION_6 ) {
             push @results,
               info(
                 IPV6_DISABLED => {
@@ -168,7 +281,7 @@ sub consistency01 {
             next;
         }
 
-        if ( not Zonemaster::Engine->config->ipv4_ok and $local_ns->address->version == $IP_VERSION_4 ) {
+        if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv4}) and $local_ns->address->version == $IP_VERSION_4 ) {
             push @results,
               info(
                 IPV4_DISABLED => {
@@ -212,6 +325,17 @@ sub consistency01 {
     } ## end foreach my $local_ns ( @{ Zonemaster::Engine::TestMethods...})
 
     my @serial_numbers = sort keys %serials;
+
+    foreach my $serial ( @serial_numbers ) {
+        push @results,
+          info(
+            SOA_SERIAL => {
+                serial  => $serial,
+                servers => join( q{;}, sort @{ $serials{$serial} } ),
+            }
+          );
+    }
+
     if ( scalar( @serial_numbers ) == 1 ) {
         push @results,
           info(
@@ -227,15 +351,6 @@ sub consistency01 {
                 count => scalar( keys %serials ),
             }
           );
-        foreach my $serial ( keys %serials ) {
-            push @results,
-              info(
-                SOA_SERIAL => {
-                    serial  => $serial,
-                    servers => join( q{;}, sort @{ $serials{$serial} } ),
-                }
-              );
-        }
         if ( $serial_numbers[-1] - $serial_numbers[0] > $MAX_SERIAL_VARIATION ) {
             push @results,
               info(
@@ -264,7 +379,7 @@ sub consistency02 {
 
         next if $nsnames_and_ip{ $local_ns->name->string . q{/} . $local_ns->address->short };
 
-        if ( not Zonemaster::Engine->config->ipv6_ok and $local_ns->address->version == $IP_VERSION_6 ) {
+        if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv6}) and $local_ns->address->version == $IP_VERSION_6 ) {
             push @results,
               info(
                 IPV6_DISABLED => {
@@ -276,7 +391,7 @@ sub consistency02 {
             next;
         }
 
-        if ( not Zonemaster::Engine->config->ipv4_ok and $local_ns->address->version == $IP_VERSION_4 ) {
+        if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv4}) and $local_ns->address->version == $IP_VERSION_4 ) {
             push @results,
               info(
                 IPV4_DISABLED => {
@@ -361,7 +476,7 @@ sub consistency03 {
 
         next if $nsnames_and_ip{ $local_ns->name->string . q{/} . $local_ns->address->short };
 
-        if ( not Zonemaster::Engine->config->ipv6_ok and $local_ns->address->version == $IP_VERSION_6 ) {
+        if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv6}) and $local_ns->address->version == $IP_VERSION_6 ) {
             push @results,
               info(
                 IPV6_DISABLED => {
@@ -373,7 +488,7 @@ sub consistency03 {
             next;
         }
 
-        if ( not Zonemaster::Engine->config->ipv4_ok and $local_ns->address->version == $IP_VERSION_4 ) {
+        if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv4}) and $local_ns->address->version == $IP_VERSION_4 ) {
             push @results,
               info(
                 IPV4_DISABLED => {
@@ -469,7 +584,7 @@ sub consistency04 {
 
         next if $nsnames_and_ip{ $local_ns->name->string . q{/} . $local_ns->address->short };
 
-        if ( not Zonemaster::Engine->config->ipv6_ok and $local_ns->address->version == $IP_VERSION_6 ) {
+        if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv6}) and $local_ns->address->version == $IP_VERSION_6 ) {
             push @results,
               info(
                 IPV6_DISABLED => {
@@ -481,7 +596,7 @@ sub consistency04 {
             next;
         }
 
-        if ( not Zonemaster::Engine->config->ipv4_ok and $local_ns->address->version == $IP_VERSION_4 ) {
+        if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv4}) and $local_ns->address->version == $IP_VERSION_4 ) {
             push @results,
               info(
                 IPV4_DISABLED => {
@@ -553,55 +668,176 @@ sub consistency04 {
     return @results;
 } ## end sub consistency04
 
+sub _get_addr_rrs {
+    my ( $class, $ns, $name, $qtype ) = @_;
+    my $p = $ns->query( $name, $qtype, { recurse => 0 } );
+    if ( !$p ) {
+        return info(
+            NO_RESPONSE => {
+                ns      => $ns->name->string,
+                address => $ns->address->short,
+            }
+        );
+    }
+    elsif ($p->is_redirect) {
+        my $p = $ns->query( $name, $qtype, { recurse => 1 } );
+        if ( $p ) {
+            return ( undef, $p->get_records_for_name( $qtype, $name, 'answer' ) );
+        } else {
+            return ( undef );
+        }
+    }
+    elsif ( $p->aa and $p->rcode eq 'NOERROR' ) {
+        return ( undef, $p->get_records_for_name( $qtype, $name, 'answer' ) );
+    }
+    elsif (not ($p->aa and $p->rcode eq 'NXDOMAIN')) {
+        return info(
+            CHILD_NS_FAILED => {
+                ns      => $ns->name->string,
+                address => $ns->address->short,
+            }
+        );
+    }
+    return ( undef );
+}
+
 sub consistency05 {
     my ( $class, $zone ) = @_;
     my @results;
 
-    my %addresses;
-    foreach my $address ( uniq map { lc( $_->address->short ) } @{ Zonemaster::Engine::TestMethods->method4( $zone ) } ) {
-        $addresses{$address} += 1;
+    my %strict_glue;
+    my %extended_glue;
+
+    # We need to work on Methods...
+    # This part of code is supposed to replace method4 call
+    my @child_nsnames;
+    my @nsnames;
+    my $ns_aref = $zone->parent->query_all( $zone->name, q{NS} );
+    my %parent_glues;
+    foreach my $p ( @{$ns_aref} ) {
+        next if not $p;
+        push @nsnames, $p->get_records_for_name( q{NS}, $zone->name );
     }
-    foreach my $address ( uniq map { lc( $_->address->short ) } @{ Zonemaster::Engine::TestMethods->method5( $zone ) } ) {
-        $addresses{$address} -= 1;
+    @child_nsnames = uniq map { name( lc( $_->nsdname ) ) } @nsnames;
+    foreach my $nsname ( @child_nsnames ) {
+        my $a_aref = $zone->parent->query_all( $nsname, q{A} );
+        my $aaaa_aref = $zone->parent->query_all( $nsname, q{AAAA} );
+        foreach my $p ( @{$a_aref} ) {
+            next if not $p;
+            foreach my $rr ( $p->get_records_for_name( q{A}, $nsname ) ) {
+                $parent_glues{ lc( $rr->owner ) . q{/} . $rr->address } = $nsname;
+            }
+        }
+        foreach my $p ( @{$aaaa_aref} ) {
+            next if not $p;
+            foreach my $rr ( $p->get_records_for_name( q{AAAA}, $nsname ) ) {
+                $parent_glues{ lc( $rr->owner ) . q{/} . $rr->address } = $nsname;
+            }
+        }
     }
 
-    my @same_address         = sort grep { $addresses{$_} == 0 } keys %addresses;
-    my @extra_address_parent = sort grep { $addresses{$_} > 0 } keys %addresses;
-    my @extra_address_child  = sort grep { $addresses{$_} < 0 } keys %addresses;
+    for my $ns_string ( keys %parent_glues ) {
+        if ( $zone->name->is_in_bailiwick( $parent_glues{$ns_string} ) ) {
+            $strict_glue{ $ns_string } = 1;
+        }
+        else {
+            push @{ $extended_glue{ $parent_glues{$ns_string} } }, $ns_string;
+        }
+    }
 
-    if ( @extra_address_parent ) {
+    my @ib_nsnames =
+      grep { $zone->name->is_in_bailiwick( $_ ) } @{ Zonemaster::Engine::TestMethods->method2and3( $zone ) };
+    my @ib_nss = grep { Zonemaster::Engine::Util::ipversion_ok( $_->address->version ) }
+      @{ Zonemaster::Engine::TestMethods->method4and5( $zone ) };
+    my %child_ib_strings;
+    for my $ib_nsname ( @ib_nsnames ) {
+        my $is_lame = 1;
+        for my $ns ( @ib_nss ) {
+            my ( $msg_a,    @rrs_a )    = $class->_get_addr_rrs( $ns, $ib_nsname, q{A} );
+            my ( $msg_aaaa, @rrs_aaaa ) = $class->_get_addr_rrs( $ns, $ib_nsname, q{AAAA} );
+
+            if ( defined $msg_a ) {
+                push @results, $msg_a;
+            }
+            if ( defined $msg_aaaa ) {
+                push @results, $msg_aaaa;
+            }
+            if ( !defined $msg_a || !defined $msg_aaaa ) {
+                $is_lame = 0;
+            }
+
+            for my $rr ( @rrs_a, @rrs_aaaa ) {
+                $child_ib_strings{ lc( $rr->name ) . q{/} . $rr->address } = 1;
+            }
+        }
+
+        if ( $is_lame ) {
+            push @results, info( CHILD_ZONE_LAME => {} );
+            return @results;
+        }
+    } ## end for my $ib_nsname ( @ib_nsnames)
+
+    my @ib_match       = grep { exists $child_ib_strings{$_} } keys %strict_glue;
+    my @ib_mismatch    = grep { !exists $child_ib_strings{$_} } keys %strict_glue;
+    my @ib_extra_child = grep { !exists $strict_glue{$_} } keys %child_ib_strings;
+
+    if ( scalar @ib_mismatch ) {
         push @results,
           info(
-            EXTRA_ADDRESS_PARENT => {
-                addresses => join( q{;}, @extra_address_parent ),
+            IN_BAILIWICK_ADDR_MISMATCH => {
+                parent_addresses => join( q{;}, sort keys %strict_glue ),
+                zone_addresses => join( q{;}, sort keys %child_ib_strings ),
             }
           );
     }
-
-    if ( @extra_address_child ) {
+    if ( scalar @ib_extra_child ) {
         push @results,
           info(
             EXTRA_ADDRESS_CHILD => {
-                addresses => join( q{;}, @extra_address_child ),
+                addresses => join( q{;}, sort @ib_extra_child ),
             }
           );
     }
 
-    if ( @extra_address_parent == 0 and @extra_address_child == 0 ) {
+    my @oob_match;
+    my @oob_mismatch;
+    for my $glue_name ( keys %extended_glue ) {
+        my @glue_strings = @{ $extended_glue{$glue_name} };
+
+        my %child_oob_strings;
+
+        my $p_a = Zonemaster::Engine->recurse( $glue_name, q{A}, q{IN} );
+        if ( $p_a ) {
+            for my $rr ( $p_a->get_records_for_name( q{A}, $glue_name, q{answer} ) ) {
+                $child_oob_strings{ lc( $rr->owner ) . q{/} . $rr->address } = 1;
+            }
+        }
+
+        my $p_aaaa = Zonemaster::Engine->recurse( $glue_name, q{AAAA}, q{IN} );
+        if ( $p_aaaa ) {
+            for my $rr ( $p_aaaa->get_records_for_name( q{AAAA}, $glue_name, q{answer} ) ) {
+                $child_oob_strings{ lc( $rr->owner ) . q{/} . $rr->address } = 1;
+            }
+        }
+
+        push @oob_match,    grep { exists $child_oob_strings{$_} } @glue_strings;
+        push @oob_mismatch, grep { !exists $child_oob_strings{$_} } @glue_strings;
+        if ( grep { !exists $child_oob_strings{$_} } @glue_strings ) {
+            push @results,
+              info(
+                OUT_OF_BAILIWICK_ADDR_MISMATCH => {
+                    parent_addresses => join( q{;}, sort @glue_strings ),
+                    zone_addresses => join( q{;}, sort keys %child_oob_strings ),
+                }
+              );
+        }
+    } ## end for my $glue_name ( keys...)
+
+    if ( !@ib_extra_child && !@ib_mismatch && !@oob_mismatch ) {
         push @results,
           info(
             ADDRESSES_MATCH => {
-                addresses => join( q{;}, @same_address ),
-            }
-          );
-    }
-
-    if ( scalar( @same_address ) == 0 ) {
-        push @results,
-          info(
-            TOTAL_ADDRESS_MISMATCH => {
-                glue  => join( q{;}, @extra_address_parent ),
-                child => join( q{;}, @extra_address_child ),
+                addresses => join( q{;}, sort @ib_match, @oob_match ),
             }
           );
     }
@@ -609,6 +845,103 @@ sub consistency05 {
     return @results;
 } ## end sub consistency05
 
+sub consistency06 {
+    my ( $class, $zone ) = @_;
+    my @results;
+    my %nsnames_and_ip;
+    my %mnames;
+    my $query_type = q{SOA};
+
+    foreach
+      my $local_ns ( @{ Zonemaster::Engine::TestMethods->method4( $zone ) }, @{ Zonemaster::Engine::TestMethods->method5( $zone ) } )
+    {
+
+        next if $nsnames_and_ip{ $local_ns->name->string . q{/} . $local_ns->address->short };
+
+        if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv6}) and $local_ns->address->version == $IP_VERSION_6 ) {
+            push @results,
+              info(
+                IPV6_DISABLED => {
+                    ns      => $local_ns->name->string,
+                    address => $local_ns->address->short,
+                    rrtype  => $query_type,
+                }
+              );
+            next;
+        }
+
+        if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv4}) and $local_ns->address->version == $IP_VERSION_4 ) {
+            push @results,
+              info(
+                IPV4_DISABLED => {
+                    ns      => $local_ns->name->string,
+                    address => $local_ns->address->short,
+                    rrtype  => $query_type,
+                }
+              );
+            next;
+        }
+
+        my $p = $local_ns->query( $zone->name, $query_type );
+
+        if ( not $p ) {
+            push @results,
+              info(
+                NO_RESPONSE => {
+                    ns      => $local_ns->name->string,
+                    address => $local_ns->address->short,
+                }
+              );
+            next;
+        }
+
+        my ( $soa ) = $p->get_records_for_name( $query_type, $zone->name );
+
+        if ( not $soa ) {
+            push @results,
+              info(
+                NO_RESPONSE_SOA_QUERY => {
+                    ns      => $local_ns->name->string,
+                    address => $local_ns->address->short,
+                }
+              );
+            next;
+        }
+        else {
+            push @{ $mnames{ lc( $soa->mname ) } }, $local_ns->name->string . q{/} . $local_ns->address->short;
+            $nsnames_and_ip{ $local_ns->name->string . q{/} . $local_ns->address->short }++;
+        }
+    } ## end foreach my $local_ns ( @{ Zonemaster::Engine::TestMethods...})
+
+    if ( scalar( keys %mnames ) == 1 ) {
+        push @results,
+          info(
+            ONE_SOA_MNAME => {
+                mname => ( keys %mnames )[0],
+            }
+          );
+    }
+    elsif ( scalar( keys %mnames ) ) {
+        push @results,
+          info(
+            MULTIPLE_SOA_MNAMES => {
+                count => scalar( keys %mnames ),
+            }
+          );
+        foreach my $mname ( keys %mnames ) {
+            push @results,
+              info(
+                SOA_MNAME => {
+                    mname   => $mname,
+                    servers => join( q{;}, @{ $mnames{$mname} } ),
+                }
+              );
+        }
+    }
+
+    return @results;
+} ## end sub consistency06
+
 1;
 
 =head1 NAME
@@ -632,9 +965,9 @@ Runs the default set of tests and returns a list of log entries made by the test
 Returns a reference to a hash, the keys of which are the names of all test methods in the module, and the corresponding values are references to
 lists with all the tags that the method can use in log entries.
 
-=item translation()
+=item tag_descriptions()
 
-Returns a refernce to a hash with translation data. Used by the builtin translation system.
+Returns a refernce to a hash with translation functions. Used by the builtin translation system.
 
 =item version()
 
@@ -666,6 +999,10 @@ Query all nameservers for NS set, and see that they have all the same content.
 
 Verify that the glue records are consistent between glue and authoritative data.
 
+=item consistency06($zone)
+
+Query all nameservers for SOA, and see that they all have the same SOA mname.
+
 =back
 
 =cut
